@@ -16,39 +16,60 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 """The tinydav WebDAV client."""
-
 from __future__ import with_statement
-from contextlib import closing
-from functools import wraps, partial
-from httplib import MULTI_STATUS, OK, CONFLICT, NO_CONTENT, UNAUTHORIZED
-from StringIO import StringIO
-from xml.etree.ElementTree import ElementTree, Element, SubElement, tostring
-from xml.parsers.expat import ExpatError
-import hashlib
-import httplib
 import sys
-import urllib
+
+PYTHON2_6 = (sys.version_info >= (2, 6))
+PYTHON2_7 = (sys.version_info >= (2, 7))
+PYTHON2 = ((2, 5) <= sys.version_info <= (3, 0))
+PYTHON3 = (sys.version_info >= (3, 0))
+
+from contextlib import closing
+from email.header import Header
+from functools import wraps, partial
+
+if PYTHON2:
+    from httplib import MULTI_STATUS, OK, CONFLICT, NO_CONTENT, UNAUTHORIZED
+    from urllib import quote as urllib_quote
+    from urllib import urlencode as urllib_urlencode
+    from StringIO import StringIO
+    import httplib
+else:
+    from http.client import MULTI_STATUS, OK, CONFLICT, NO_CONTENT
+    from http.client import UNAUTHORIZED
+    from io import BytesIO
+    from io import StringIO
+    from urllib.parse import quote as urllib_quote
+    from urllib.parse import urlencode as urllib_urlencode
+    import base64
+    import http.client as httplib
+
+from xml.etree.ElementTree import ElementTree, Element, SubElement, tostring
+
+if PYTHON2_7 or PYTHON3:
+    from xml.etree.ElementTree import ParseError
+else:
+    from xml.parsers.expat import ExpatError as ParseError
+
+import hashlib
 
 from tinydav import creator, util
 from tinydav.exception import HTTPError, HTTPUserError, HTTPServerError
 
 __author__ = "Manuel Hermann <manuel-hermann@gmx.net>"
 __license__ = "LGPL"
-__version__ = "0.6.8"
+__version__ = "0.7.0"
 
 __all__ = (
     "HTTPError", "HTTPUserError", "HTTPServerError",
     "HTTPClient", "WebDAVClient",
 )
 
-PYTHON2_6 = (sys.version_info >= (2, 6))
-PYTHON2_7 = (sys.version_info >= (2, 7))
-
 # RFC 2518, 9.8 Timeout Request Header
 # The timeout value for TimeType "Second" MUST NOT be greater than 2^32-1.
 MAX_TIMEOUT = 2**32-1
 
-ACTIVELOCK = "/{DAV:}lockdiscovery/{DAV:}activelock"
+ACTIVELOCK = "./{DAV:}lockdiscovery/{DAV:}activelock"
 
 # map with default ports mapped to http protocol
 PROTOCOL = {
@@ -59,6 +80,10 @@ PROTOCOL = {
 }
 
 
+default_header_encoding = "utf-8"
+separate_query_sequences = True
+
+
 # Responses
 class HTTPResponse(int):
     """Result from HTTP request.
@@ -100,7 +125,10 @@ class HTTPResponse(int):
 
     def __repr__(self):
         """Return representation."""
-        return "<%s: %d>" % (self.__class__.__name__, self)
+        if PYTHON2:
+            return "<%s: %d>" % (self.__class__.__name__, self)
+        else:
+            return "<{0}: {1}>".format(self.__class__.__name__, self)
 
     def __str__(self):
         """Return string representation."""
@@ -111,9 +139,13 @@ class HTTPResponse(int):
         auth = util.parse_authenticate(value)
         for attrname in ("schema", "realm", "domain", "nonce", "opaque"):
             setattr(self, attrname, auth.get(attrname))
-        stale = auth.get("stale", "false")
+        stale = auth.get("stale")
+        if stale is None:
+            stale = "false"
         self.stale = (stale.lower() == "true")
-        algorithm = auth.get("algorithm", "MD5")
+        algorithm = auth.get("algorithm")
+        if algorithm is None:
+            algorithm = "MD5"
         self.algorithm = getattr(hashlib, algorithm.lower())
 
 
@@ -165,7 +197,7 @@ class WebDAVResponse(HTTPResponse):
         if self.is_multistatus:
             # RFC 2518, 12.9 multistatus XML Element
             # <!ELEMENT multistatus (response+, responsedescription?) >
-            return len(self._etree.findall("/{DAV:}response"))
+            return len(self._etree.findall("./{DAV:}response"))
         return 1
 
     def __iter__(self):
@@ -179,7 +211,7 @@ class WebDAVResponse(HTTPResponse):
         if self.is_multistatus:
             # RFC 2518, 12.9 multistatus XML Element
             # <!ELEMENT multistatus (response+, responsedescription?) >
-            for response in self._etree.findall("/{DAV:}response"):
+            for response in self._etree.findall("./{DAV:}response"):
                 yield MultiStatusResponse(response)
         else:
             yield self
@@ -192,10 +224,15 @@ class WebDAVResponse(HTTPResponse):
 
         """
         try:
-            parse_me = StringIO(self.content)
+            if PYTHON2:
+                parse_me = StringIO(self.content)
+            else:
+                parse_me = BytesIO(self.content)
             self._etree.parse(parse_me)
-        except ExpatError, e:
-            self.parse_error = e
+        except ParseError:
+            # get the exception object this way to be compatible with Python
+            # versions 2.5 up to 3.x
+            self.parse_error = sys.exc_info()[1]
             # don't fail on further processing
             self._etree.parse(StringIO("<root><empty/></root>"))
 
@@ -466,26 +503,41 @@ class MultiStatusResponse(int):
             raise KeyError(name)
         return prop
 
-    def __iter__(self):
-        """Iterator over propertynames with their namespaces."""
-        return self.iterkeys()
+    if PYTHON2:
+        def __iter__(self):
+            """Iterator over propertynames with their namespaces."""
+            return self.iterkeys()
+    else:
+        def __iter__(self):
+            """Iterator over propertynames with their namespaces."""
+            return self.keys()
 
-    def keys(self):
-        """Return list of propertynames with their namespaces.
+    if PYTHON2:
+        def keys(self):
+            """Return list of propertynames with their namespaces.
 
-        No namespaces for DAV properties.
+            No namespaces for DAV properties.
 
-        """
-        return list(self.iterkeys())
+            """
+            return list(self.iterkeys())
 
-    def iterkeys(self, cut_dav_ns=True):
-        """Iterate over propertynames with their namespaces.
+        def iterkeys(self, cut_dav_ns=True):
+            """Iterate over propertynames with their namespaces.
 
-        cut_dav_ns -- No namespaces for DAV properties when this is True.
+            cut_dav_ns -- No namespaces for DAV properties when this is True.
 
-        """
-        for (tagname, value) in self.iteritems(cut_dav_ns):
-            yield tagname
+            """
+            for (tagname, value) in self.iteritems(cut_dav_ns):
+                yield tagname
+    else:
+        def keys(self, cut_dav_ns=True):
+            """Iterate over propertynames with their namespaces.
+
+            cut_dav_ns -- No namespaces for DAV properties when this is True.
+
+            """
+            for (tagname, value) in self.items(cut_dav_ns):
+                yield tagname
 
     def items(self):
         """Return list of 2-tuples with propertyname and ElementTree element."""
@@ -506,6 +558,10 @@ class MultiStatusResponse(int):
                 tagname = tagname[6:]
             yield (tagname, prop)
 
+    if PYTHON3:
+        items = iteritems
+        del iteritems
+
     def get(self, key, default=None, namespace=None):
         """Return value for requested property.
 
@@ -543,14 +599,24 @@ class MultiStatusResponse(int):
             self._href = self.response.findtext("{DAV:}href")
         return self._href
 
-    @property
-    def namespaces(self):
-        """Return frozenset of namespaces."""
-        if self._namespaces is None:
-            self._namespaces = frozenset(util.extract_namespace(key)
-                                         for key in self.iterkeys(False)
-                                         if util.extract_namespace(key))
-        return self._namespaces
+    if PYTHON2:
+        @property
+        def namespaces(self):
+            """Return frozenset of namespaces."""
+            if self._namespaces is None:
+                self._namespaces = frozenset(util.extract_namespace(key)
+                                             for key in self.iterkeys(False)
+                                             if util.extract_namespace(key))
+            return self._namespaces
+    else:
+        @property
+        def namespaces(self):
+            """Return frozenset of namespaces."""
+            if self._namespaces is None:
+                self._namespaces = frozenset(util.extract_namespace(key)
+                                             for key in self.keys(False)
+                                             if util.extract_namespace(key))
+            return self._namespaces
 
 
 # Clients
@@ -610,8 +676,11 @@ class HTTPClient(object):
         self.strict = strict
         self.timeout = timeout
         self.source_address = source_address
-        self.key_file = None
-        self.cert_file = None
+        if PYTHON2:
+            self.key_file = None
+            self.cert_file = None
+        else:
+            self.context = None
         self.headers = dict()
         self.cookie = None
         self._do_digest_auth = False
@@ -627,8 +696,11 @@ class HTTPClient(object):
         if self.protocol == "http":
             return httplib.HTTPConnection(*args, **kwargs)
         # setup HTTPS
-        kwargs["key_file"] = self.key_file
-        kwargs["cert_file"] = self.cert_file
+        if PYTHON2:
+            kwargs["key_file"] = self.key_file
+            kwargs["cert_file"] = self.cert_file
+        else:
+            kwargs["context"] = self.context
         return httplib.HTTPSConnection(*args, **kwargs)
 
     def _request(self, method, uri, content=None, headers=None):
@@ -674,38 +746,67 @@ class HTTPClient(object):
         The headers will contain the authorization headers, if given.
 
         uri -- URI the request is for.
-        headers -- Mapping with additional headers to send.
+        headers -- Mapping with additional headers to send. Unicode values that
+                   are no ASCII will be MIME-encoded with UTF-8. Set 
+                   tinydav.default_header_encoding to another encoding, if
+                   UTF-8 doesn't suit you.
         query -- Mapping with key/value-pairs to be added as query to the URI.
 
         """
-        uri = urllib.quote(uri)
+        uri = urllib_quote(uri)
         # collect headers
         sendheaders = dict(self.headers)
         if headers:
             sendheaders.update(headers)
+        for (key, value) in sendheaders.items():
+            if key.lower() != "authorization":
+                value = str(Header(value, default_header_encoding))
+            sendheaders[key] = value
         # construct query string
         if query:
-            querystr = urllib.urlencode(query)
+            querystr = urllib_urlencode(query, doseq=separate_query_sequences)
             uri = "%s?%s" % (uri, querystr)
         return (uri, sendheaders)
 
-    def setbasicauth(self, user, password):
-        """Set authorization header for basic auth.
+    if PYTHON2:
+        def setbasicauth(self, user, password):
+            """Set authorization header for basic auth.
 
-        user -- Username
-        password -- Password for user.
+            user -- Username
+            password -- Password for user.
 
-        """
-        # RFC 2068, 11.1 Basic Authentication Scheme
-        # basic-credentials = "Basic" SP basic-cookie
-        # basic-cookie   = <base64 [7] encoding of user-pass,
-        # except not limited to 76 char/line>
-        # user-pass   = userid ":" password
-        # userid      = *<TEXT excluding ":">
-        # password    = *TEXT
-        userpw = "%s:%s" % (user, password)
-        auth = userpw.encode("base64").rstrip()
-        self.headers["Authorization"] = "Basic %s" % auth
+            """
+            # RFC 2068, 11.1 Basic Authentication Scheme
+            # basic-credentials = "Basic" SP basic-cookie
+            # basic-cookie   = <base64 [7] encoding of user-pass,
+            # except not limited to 76 char/line>
+            # user-pass   = userid ":" password
+            # userid      = *<TEXT excluding ":">
+            # password    = *TEXT
+            userpw = "%s:%s" % (user, password)
+            auth = userpw.encode("base64").rstrip()
+            self.headers["Authorization"] = "Basic %s" % auth
+    else:
+        def setbasicauth(self, user, password,
+                         b64encoder=base64.standard_b64encode):
+            """Set authorization header for basic auth.
+
+            user -- Username as bytes string.
+            password -- Password for user as bytes.
+            encoder -- Base64 encoder function. Default is the standard 
+                       encoder. Should not be changed.
+
+            """
+            # RFC 2068, 11.1 Basic Authentication Scheme
+            # basic-credentials = "Basic" SP basic-cookie
+            # basic-cookie   = <base64 [7] encoding of user-pass,
+            # except not limited to 76 char/line>
+            # user-pass   = userid ":" password
+            # userid      = *<TEXT excluding ":">
+            # password    = *TEXT
+            userpw = user + bytes(":", "ascii") + password
+            auth = b64encoder(userpw).decode("ascii")
+            self.headers["Authorization"] = "Basic {0}".format(auth)
 
     def setcookie(self, cookie):
         """Set cookie class to be used in requests.
@@ -715,21 +816,35 @@ class HTTPClient(object):
         """
         self.cookie = cookie
 
-    def setssl(self, key_file=None, cert_file=None):
-        """Set SSL key file and/or certificate chain file for HTTPS.
+    if PYTHON2:
+        def setssl(self, key_file=None, cert_file=None):
+            """Set SSL key file and/or certificate chain file for HTTPS.
 
-        Calling this method has the side effect of setting the protocol to
-        https.
+            Calling this method has the side effect of setting the protocol to
+            https.
 
-        key_file -- The name of a PEM formatted file that contains your
-                    private key.
-        cert_file -- PEM formatted certificate chain file (see Python doc for
-                     httplib).
-        """
-        self.key_file = key_file
-        self.cert_file = cert_file
-        if any((key_file, cert_file)):
+            key_file -- The name of a PEM formatted file that contains your
+                        private key.
+            cert_file -- PEM formatted certificate chain file (see Python doc
+                         for httplib).
+            """
+            self.key_file = key_file
+            self.cert_file = cert_file
+            if any((key_file, cert_file)):
+                self.protocol = "https"
+    else:
+        def setssl(self, context):
+            """Set SSLContext for this connection.
+
+            Calling this method has the side effect of setting the protocol to
+            https.
+
+            context -- ssl.SSLContext instance describing the various SSL
+                       options.
+
+            """
             self.protocol = "https"
+            self.context = context
 
     def options(self, uri, headers=None):
         """Make OPTIONS request and return status.
@@ -790,8 +905,8 @@ class HTTPClient(object):
                         the actual value (or file-like object) and an encoding
                         for this value (or the content-type in case of a 
                         file-like object).
-        encoding -- Send multipart content encoding with this encoding. Default
-                    is ASCII.
+        encoding -- Send multipart content with this encoding. Default is
+                    ASCII.
         with_filenames -- If True, a multipart's files will be sent with the
                           filename paramenter set. Default is False.
 
@@ -808,7 +923,7 @@ class HTTPClient(object):
                 headers.update(multihead)
             else:
                 headers["content-type"] = "application/x-www-form-urlencoded"
-                content = urllib.urlencode(content)
+                content = urllib_urlencode(content)
         if hasattr(content, "read") and not PYTHON2_6:
             # python 2.5 httlib cannot handle file-like objects
             content = content.read()
@@ -999,8 +1114,7 @@ class CoreWebDAVClient(HTTPClient):
         depth = util.get_depth(depth)
         # check mutually exclusive arguments
         if all([properties, include]):
-            raise ValueError("properties and include are "
-                                 "mutually exclusive")
+            raise ValueError("properties and include are mutually exclusive")
         (uri, headers) = self._prepare(uri, headers)
         # additional headers needed for PROPFIND
         headers["Depth"] = depth
